/*
 *
 * @APPLE_LICENSE_HEADER_START@
 *
 * Portions Copyright (c) 1999-2007 Apple Inc.  All Rights Reserved.
 *
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 *
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 *
 * @APPLE_LICENSE_HEADER_END@
 */
//
//  HIDDescriptors.c
//  HIDReportTracer
//
//  Copied from Apple's USB Prober 2 project (DecodeHIDDescriptor.m)
//  Additions from Apple's "HID Calibrator" sample code (HID_usage_strings.plist)
//

#include <stdint.h>
#include <stdlib.h>
#include <IOKit/hid/IOHIDUsageTables.h>
#include "HIDDescriptors.h"

const char* usage_page_name(uint32_t usagePage) {
    
    if (usagePage >= 0xff00 && usagePage <= 0xffff)
        return "Vendor-defined";
    
    switch (usagePage)
    {
        case kUsage_PageGenericDesktop: return "Generic Desktop";
        case kUsage_PageSimulationControls: return "Simulation Controls";
        case kUsage_PageVRControls: return "VR Controls";
        case kUsage_PageSportControls: return "Sports Controls";
        case kUsage_PageGameControls: return "Game Controls";
        case 0x6: return "Generic Device Controls";
        case kUsage_PageKeyboard: return "Keyboard/Keypad";
            
        case kUsage_PageLED: return "LED";
        case kUsage_PageButton: return "Button";
        case kUsage_PageOrdinal: return "Ordinal";
        case kUsage_PageTelephonyDevice: return "Telephony Device";
        case kUsage_PageConsumer: return "Consumer";
        case kUsage_PageDigitizers: return "Digitizer";
        case kUsage_PagePID: return "PID";
        case kUsage_PageUnicode: return "Unicode";
        case kUsage_PageAlphanumericDisplay: return "Alphanumeric Display";
        case 0x40: return "Medical Instruments";
        case kUsage_PageMonitor: return "Monitor";
        case kUsage_PageMonitorEnumeratedValues: return "Monitor Enumerated Values";
        case kUsage_PageMonitorVirtualControl: return "VESA Virtual Controls";
        case kUsage_PageMonitorReserved: return "Monitor Class reserved";
        case kUsage_PagePowerDevice: return "Power Device";
        case kUsage_PageBatterySystem: return "Battery System";
        case kUsage_PowerClassReserved: return "Power Class reserved";
        case kUsage_PowerClassReserved2: return "Power Class reserved";
        case 0x8c: return "Bar Code Scanner";
        case 0x8d: return "Scale";
        case 0x8e: return "Magnetic Stripe Reading (MSR) Devices";
        case 0x8f: return "Reserved Point of Sale";
        case 0x90: return "Camera Control";
        case 0x91: return "Arcade";

        default: return NULL;
    }
}

const char* usage_name(uint32_t page, uint32_t usage) {
    if (page == kUsage_PageGenericDesktop)
    {
        switch (usage)
        {
            case 0x0: return "Undefined Usage";
            case kUsage_01_Pointer: return "Pointer";
            case kUsage_01_Mouse: return "Mouse";
            case kUsage_01_Joystick: return "Joystick";
            case kUsage_01_GamePad: return "GamePad";
            case kUsage_01_Keyboard: return "Keyboard";
            case kUsage_01_Keypad: return "Keypad";
                
            case 0x08: return "Multi-axis Controller";
            case 0x09: return "Tablet PC System Controls";
                
            case kUsage_01_X: return "X";
            case kUsage_01_Y: return "Y";
            case kUsage_01_Z: return "Z";
            case kUsage_01_Rx: return "Rx";
            case kUsage_01_Ry: return "Ry";
            case kUsage_01_Rz: return "Rz";
            case kUsage_01_Slider: return "Slider";
            case kUsage_01_Dial: return "Dial";
            case kUsage_01_Wheel: return "Wheel";
            case kUsage_01_HatSwitch: return "Hat Switch";
            case kUsage_01_CountedBuffer: return "Counted Buffer";
            case kUsage_01_ByteCount: return "Byte Count";
            case kUsage_01_MotionWakeup: return "Motion Wakeup";
                
            case 0x3d: return "Start";
            case 0x3e: return "Select";
                
            case kUsage_01_Vx: return "Vx";
            case kUsage_01_Vy: return "Vy";
            case kUsage_01_Vz: return "Vz";
            case kUsage_01_Vbrx: return "Vbrx";
            case kUsage_01_Vbry: return "Vbry";
            case kUsage_01_Vbrz: return "Vbrz";
            case kUsage_01_Vno: return "Vno";
                
            case 0x47: return "Feature Notification";
            case 0x48: return "Resolution Multiplier";
                
            case kUsage_01_SystemControl: return "System Control";
            case kUsage_01_SystemPowerDown: return "System Power Down";
            case kUsage_01_SystemSleep: return "System Sleep";
            case kUsage_01_SystemWakeup: return "System Wakeup";
            case kUsage_01_SystemContextMenu: return "System Context Menu";
            case kUsage_01_SystemMainMenu: return "System Main Menu";
            case kUsage_01_SystemAppMenu: return "System App Menu";
            case kUsage_01_SystemMenuHelp: return "System Menu Help";
            case kUsage_01_SystemMenuExit: return "System Menu Exit";
            case kUsage_01_SystemMenuSelect: return "System Menu Select";
            case kUsage_01_SystemMenuRight: return "System Menu Right";
            case kUsage_01_SystemMenuLeft: return "System Menu Left";
            case kUsage_01_SystemMenuUp: return "System Menu Up";
            case kUsage_01_SystemMenuDown: return "System Menu Down";
                
            case 0x8e: return "System Cold Restart";
            case 0x8f: return "System Warm Restart";
                
            case 0x0090: return "DPad Up";
            case 0x0091: return "DPad Down";
            case 0x0092: return "DPad Right";
            case 0x0093: return "DPad Left";
                
                // A few more...
                
            default: return NULL;
        }
    }
    else if (page == kUsage_PageSimulationControls)
    {
        switch (usage)
        {
            case 0x0000: return "Undefined Usage";
            case 0x0001: return "Flight Simulation Device";
            case 0x0002: return "Automobile Simulation Device";
            case 0x0003: return "Tank Simulation Device";
            case 0x0004: return "Spaceship Simulation Device";
            case 0x0005: return "Submarine Simulation Device";
            case 0x0006: return "Sailing Simulation Device";
            case 0x0007: return "Motorcycle Simulation Device";
            case 0x0008: return "Sports Simulation Device";
            case 0x0009: return "Airplane Simulation Device";
            case 0x000A: return "Helicopter Simulation Device";
            case 0x000B: return "Magic Carpet Simulation Device";
            case 0x000C: return "Bicycle Simulation Device";
            case 0x0020: return "Flight Control Stick";
            case 0x0021: return "Flight Stick";
            case 0x0022: return "Cyclic Control";
            case 0x0023: return "Cyclic Trim";
            case 0x0024: return "Flight Yoke";
            case 0x0025: return "Track Control";
            case 0x00B0: return "Aileron";
            case 0x00B1: return "Aileron Trim";
            case 0x00B2: return "Anti Torque Control";
            case 0x00B5: return "Collective Control";
            case 0x00B6: return "Dive Brake";
            case 0x00B7: return "Electronic Countermeasures";
            case 0x00B8: return "Elevator";
            case 0x00B9: return "Elevator Trim";
            case 0x00BA: return "Rudder";
            case 0x00BB: return "Throttle";
            case 0x00BC: return "Flight Communications";
            case 0x00BD: return "Flare Release";
            case 0x00BE: return "Landing Gear";
            case 0x00BF: return "Toe Brake";
            case 0x00C0: return "Trigger";
            case 0x00C1: return "Weapons Arm";
            case 0x00C2: return "Weapons";
            case 0x00C3: return "Wing Flaps";
            case 0x00C4: return "Accelerator";
            case 0x00C5: return "Brake";
            case 0x00C6: return "Clutch";
            case 0x00C7: return "Shifter";
            case 0x00C8: return "Steering";
            case 0x00C9: return "Turret Direction";
            case 0x00CA: return "Barrel Elevation";
            case 0x00CB: return "Dive Plane";
            case 0x00CC: return "Ballast";
            case 0x00CD: return "Bicycle Crank";
            case 0x00CE: return "Handle Bars";
            case 0x00CF: return "Front Brake";
            case 0x00D0: return "Rear Brake";
            default: return NULL;
        }
    }
    else if (page == kUsage_PageVRControls)
    {
        switch (usage) 
        {
            case 0x0000: return "Undefined Usage";
            case 0x0001: return "Belt";
            case 0x0002: return "Body Suit";
            case 0x0003: return "Flexor";
            case 0x0004: return "Glove";
            case 0x0005: return "Head Tracker";
            case 0x0006: return "Head Mounted Display";
            case 0x0007: return "Hand Tracker";
            case 0x0008: return "Oculometer";
            case 0x0009: return "Vest";
            case 0x000A: return "Animatronic Device";
            case 0x0020: return "Stereo Enable";
            case 0x0021: return "Display Enable";
            default: return NULL;
        }
    }
    else if (page == kUsage_PageSportControls)
    {
        switch (usage)
        {
            case 0x0000: return "Undefined Usage";
            case 0x0001: return "Baseball Bat";
            case 0x0002: return "Golf Club";
            case 0x0003: return "Rowing Machine";
            case 0x0004: return "Treadmill";
            case 0x0030: return "Oar";
            case 0x0031: return "Slope";
            case 0x0032: return "Rate";
            case 0x0033: return "Stick Speed";
            case 0x0034: return "Stick Face Angle";
            case 0x0035: return "Stick Heel Or Toe";
            case 0x0036: return "Stick Follow Through";
            case 0x0037: return "Stick Tempo";
            case 0x0038: return "Stick Type";
            case 0x0039: return "Stick Height";
            case 0x0050: return "Putter";
            case 0x0051: return "1 Iron";
            case 0x0052: return "2 Iron";
            case 0x0053: return "3 Iron";
            case 0x0054: return "4 Iron";
            case 0x0055: return "5 Iron";
            case 0x0056: return "6 Iron";
            case 0x0057: return "7 Iron";
            case 0x0058: return "8 Iron";
            case 0x0059: return "9 Iron";
            case 0x005A: return "10 Iron";
            case 0x005B: return "11 Iron";
            case 0x005C: return "Sand Wedge";
            case 0x005D: return "Loft Wedge";
            case 0x005E: return "Power Wedge";
            case 0x005F: return "1 Wood";
            case 0x0060: return "3 Wood";
            case 0x0061: return "5 Wood";
            case 0x0062: return "7 Wood";
            case 0x0063: return "9 Wood";
            default: return NULL;
        }
    }
    else if (page == kUsage_PageGameControls)
    {
        switch (usage)
        {
            case 0x0000: return "Undefined Usage";
            case 0x0001: return "3D Game Controller";
            case 0x0002: return "Pinball Device";
            case 0x0003: return "Gun Device";
            case 0x0020: return "Point of View";
            case 0x0021: return "Turn Right Or Left";
            case 0x0022: return "Pitch Up Or Down";
            case 0x0023: return "Roll Right Or Left";
            case 0x0024: return "Move Right Or Left";
            case 0x0025: return "Move Forward Or Backward";
            case 0x0026: return "Move Up Or Down";
            case 0x0027: return "Lean Right Or Left";
            case 0x0028: return "Lean Forward Or Backward";
            case 0x0029: return "Height Of POV";
            case 0x002A: return "Flipper";
            case 0x002B: return "Secondary Flipper";
            case 0x002C: return "Bump";
            case 0x002D: return "New Game";
            case 0x002E: return "Shoot Ball";
            case 0x002F: return "Player";
            case 0x0030: return "Gun Bolt";
            case 0x0031: return "Gun Clip";
            case 0x0032: return "Gun Selector";
            case 0x0033: return "Gun Single Shot";
            case 0x0034: return "Gun Burst";
            case 0x0035: return "Gun Automatic";
            case 0x0036: return "Gun Safety";
            case 0x0037: return "Gamepad Fire Or Jump";
            case 0x0039: return "Gamepad Trigger";
            default: return NULL;
        }
    }
    else if (page == kUsage_PageKeyboard)
    {
        switch (usage)
        {
            case 0x0000: return "Undefined Usage";
            case 0x0001: return "Error Roll Over";
            case 0x0002: return "POST Fail";
            case 0x0003: return "Error Undefined";
            case 0x0004: return "A";
            case 0x0005: return "B";
            case 0x0006: return "C";
            case 0x0007: return "D";
            case 0x0008: return "E";
            case 0x0009: return "F";
            case 0x000A: return "G";
            case 0x000B: return "H";
            case 0x000C: return "I";
            case 0x000D: return "J";
            case 0x000E: return "K";
            case 0x000F: return "L";
            case 0x0010: return "M";
            case 0x0011: return "N";
            case 0x0012: return "O";
            case 0x0013: return "P";
            case 0x0014: return "Q";
            case 0x0015: return "R";
            case 0x0016: return "S";
            case 0x0017: return "T";
            case 0x0018: return "U";
            case 0x0019: return "V";
            case 0x001A: return "W";
            case 0x001B: return "X";
            case 0x001C: return "Y";
            case 0x001D: return "Z";
            case 0x001E: return "1";
            case 0x001F: return "2";
            case 0x0020: return "3";
            case 0x0021: return "4";
            case 0x0022: return "5";
            case 0x0023: return "6";
            case 0x0024: return "7";
            case 0x0025: return "8";
            case 0x0026: return "9";
            case 0x0027: return "0";
            case 0x0028: return "Return";
            case 0x0029: return "Escape";
            case 0x002A: return "Delete";
            case 0x002B: return "Tab";
            case 0x002C: return "Spacebar";
            case 0x002D: return "Dash";
            case 0x002E: return "Equal";
            case 0x002F: return "Left Square Bracket";
            case 0x0030: return "Right Square Bracket";
            case 0x0031: return "Slash";
            case 0x0032: return "Non-US #";
            case 0x0033: return "Semi-Colan";
            case 0x0034: return "Single Quote";
            case 0x0035: return "Grave Accent";
            case 0x0036: return "Comma";
            case 0x0037: return "Period";
            case 0x0038: return "Slash";
            case 0x0039: return "Caps Lock";
            case 0x003A: return "F1";
            case 0x003B: return "F2";
            case 0x003C: return "F3";
            case 0x003D: return "F4";
            case 0x003E: return "F5";
            case 0x003F: return "F6";
            case 0x0040: return "F7";
            case 0x0041: return "F8";
            case 0x0042: return "F9";
            case 0x0043: return "F10";
            case 0x0044: return "F11";
            case 0x0045: return "F12";
            case 0x0046: return "Print Screen";
            case 0x0047: return "Scroll Lock";
            case 0x0048: return "Pause";
            case 0x0049: return "Insert";
            case 0x004A: return "Home";
            case 0x004B: return "Page Up";
            case 0x004C: return "Delete Forward";
            case 0x004D: return "End";
            case 0x004E: return "Page Down";
            case 0x004F: return "Right Arrow";
            case 0x0050: return "Left Arrow";
            case 0x0051: return "Down Arrow";
            case 0x0052: return "Up Arrow";
            case 0x0053: return "Keypad NumLock";
            case 0x0054: return "Keypad Slash";
            case 0x0055: return "Keypad Asterisk";
            case 0x0056: return "Keypad Dash";
            case 0x0057: return "Keypad Plus";
            case 0x0058: return "Keypad Enter";
            case 0x0059: return "Keypad 1";
            case 0x005A: return "Keypad 2";
            case 0x005B: return "Keypad 3";
            case 0x005C: return "Keypad 4";
            case 0x005D: return "Keypad 5";
            case 0x005E: return "Keypad 6";
            case 0x005F: return "Keypad 7";
            case 0x0060: return "Keypad 8";
            case 0x0061: return "Keypad 9";
            case 0x0062: return "Keypad 0";
            case 0x0063: return "Keypad Period";
            case 0x0064: return "Non-US Backslash";
            case 0x0065: return " Application ";
            case 0x0066: return " Power ";
            case 0x0067: return "Keypad Equal";
            case 0x0068: return "F13";
            case 0x0069: return "F14";
            case 0x006A: return "F15";
            case 0x006B: return "F16";
            case 0x006C: return "F17";
            case 0x006D: return "F18";
            case 0x006E: return "F19";
            case 0x006F: return "F20";
            case 0x0070: return "F21";
            case 0x0071: return "F22";
            case 0x0072: return "F23";
            case 0x0073: return "F24";
            case 0x0074: return "Execute";
            case 0x0075: return "Help";
            case 0x0076: return "Menu";
            case 0x0077: return "Select";
            case 0x0078: return "Stop";
            case 0x0079: return "Again";
            case 0x007A: return "Undo";
            case 0x007B: return "Cut";
            case 0x007C: return "Copy";
            case 0x007D: return "Paste";
            case 0x007E: return "Find";
            case 0x007F: return "Mute";
            case 0x0080: return "Volume Up";
            case 0x0081: return "Volume Down";
            case 0x0082: return "Locking Caps Lock";
            case 0x0083: return "Locking Num Lock";
            case 0x0084: return "Locking Scroll Lock";
            case 0x0085: return "Keypad Comma";
            case 0x0086: return "Keypad Equal Sign for AS-400";
            case 0x0087: return "International1";
            case 0x0088: return "International2";
            case 0x0089: return "International3";
            case 0x008A: return "International4";
            case 0x008B: return "International5";
            case 0x008C: return "International6";
            case 0x008D: return "International7";
            case 0x008E: return "International8";
            case 0x008F: return "International9";
            case 0x0090: return "LANG1";
            case 0x0091: return "LANG2";
            case 0x0092: return "LANG3";
            case 0x0093: return "LANG4";
            case 0x0094: return "LANG5";
            case 0x0095: return "LANG6";
            case 0x0096: return "LANG7";
            case 0x0097: return "LANG8";
            case 0x0098: return "LANG9";
            case 0x0099: return "Alternate Erase";
            case 0x009A: return "SysReq or Attention";
            case 0x009B: return "Cancel";
            case 0x009C: return "Clear";
            case 0x009D: return "Prior";
            case 0x009E: return "Return";
            case 0x009F: return "Keyboard Separator";
            case 0x00A0: return "Out";
            case 0x00A1: return "Oper";
            case 0x00A2: return "Clear or Again";
            case 0x00A3: return "CrSel or Props";
            case 0x00A4: return "ExSel";
                // A few more...
            case 0x00E0: return "Left Control";
            case 0x00E1: return "Left Shift";
            case 0x00E2: return "Left Alt";
            case 0x00E3: return "Left GUI";
            case 0x00E4: return "Right Control";
            case 0x00E5: return "Right Shift";
            case 0x00E6: return "Right Alt";
            case 0x00E7: return "Right GUI";

            default: return NULL;
        }
    }
    else if (page == kUsage_PageLED)
    {
        switch (usage)
        {
            case 0x0000: return "Undefined Usage";
            case 0x0001: return "Num Lock";
            case 0x0002: return "Caps Lock";
            case 0x0003: return "Scroll Lock";
            case 0x0004: return "Compose";
            case 0x0005: return "Kana";
            case 0x0006: return "Power";
            case 0x0007: return "Shift";
            case 0x0008: return "DoNotDisturb";
            case 0x0009: return "Mute";
            case 0x000A: return "Tone Enable";
            case 0x000B: return "High Cut Filter";
            case 0x000C: return "Low Cut Filter";
            case 0x000D: return "Equalizer Enable";
            case 0x000E: return "Sound Field On";
            case 0x000F: return "Surround On";
            case 0x0010: return "Repeat";
            case 0x0011: return "Stereo";
            case 0x0012: return "Sampling Rate Detect";
            case 0x0013: return "Spinning";
            case 0x0014: return "CAV";
            case 0x0015: return "CLV";
            case 0x0016: return "Recording Format Detect";
            case 0x0017: return "Off Hook";
            case 0x0018: return "Ring";
            case 0x0019: return "Message Waiting";
            case 0x001A: return "Data Mode";
            case 0x001B: return "Battery Operation";
            case 0x001C: return "Battery OK";
            case 0x001D: return "Battery Low";
            case 0x001E: return "Speaker";
            case 0x001F: return "Headset";
            case 0x0020: return "Hold";
            case 0x0021: return "Microphone";
            case 0x0022: return "Coverage";
            case 0x0023: return "Night Mode";
            case 0x0024: return "Send Calls";
            case 0x0025: return "Call Pickup";
            case 0x0026: return "Conference";
            case 0x0027: return "Stand By";
            case 0x0028: return "Camera On";
            case 0x0029: return "Camera Off";
            case 0x002A: return "On Line";
            case 0x002B: return "Off Line";
            case 0x002C: return "Busy";
            case 0x002D: return "Ready";
            case 0x002E: return "Paper Out";
            case 0x002F: return "Paper Jam";
            case 0x0030: return "Remote";
            case 0x0031: return "Forward";
            case 0x0032: return "Reverse";
            case 0x0033: return "Stop";
            case 0x0034: return "Rewind";
            case 0x0035: return "Fast Forward";
            case 0x0036: return "Play";
            case 0x0037: return "Pause";
            case 0x0038: return "Record";
            case 0x0039: return "Error";
            case 0x003A: return "Usage Selected Indicator";
            case 0x003B: return "Usage In Use Indicator";
            case 0x003C: return "Usage Multi Mode indicator";
            case 0x003D: return "Indicator On";
            case 0x003E: return "Indicator Flash";
            case 0x003F: return "Indicator Slow Blink";
            case 0x0040: return "Indicator Fast Blink";
            case 0x0041: return "Indicator Off";
            case 0x0042: return "Flash On Time";
            case 0x0043: return "Slow Blink On Time";
            case 0x0044: return "Slow Blink Off Time";
            case 0x0045: return "Fast Blink On Time";
            case 0x0046: return "Fast Blink Off Time";
            case 0x0047: return "Usage Indicator Color";
            case 0x0048: return "Indicator Red";
            case 0x0049: return "Indicator Green";
            case 0x004A: return "Indicator Amber";
            case 0x004B: return "Generic Indicator";
            case 0x004C: return "System Suspend";
            case 0x004D: return "External Power Connected";
            default: return NULL;
        }
    }
    else if (page == kUsage_PageConsumer)
    {
        switch (usage)
        {
            case 0x0000: return "Undefined Usage";
            case 0x0001: return "Consumer Control";
            case 0x0002: return "Numeric KeyPad";
            case 0x0003: return "Programmable Buttons";
            case 0x0020: return "Plus 10";
            case 0x0021: return "Plus 100";
            case 0x0022: return "AM Or PM";
            case 0x0030: return "Power";
            case 0x0031: return "Reset";
            case 0x0032: return "Sleep";
            case 0x0033: return "Sleep After";
            case 0x0034: return "Sleep Mode";
            case 0x0035: return "Illumination";
            case 0x0036: return "Function Buttons";
            case 0x0040: return "Menu";
            case 0x0041: return "Menu Pick";
            case 0x0042: return "Menu Up";
            case 0x0043: return "Menu Down";
            case 0x0044: return "Menu Left";
            case 0x0045: return "Menu Right";
            case 0x0046: return "Menu Escape";
            case 0x0047: return "Menu Value Increase";
            case 0x0048: return "Menu Value Decrease";
            case 0x0060: return "Data On Screen";
            case 0x0061: return "Closed Caption";
            case 0x0062: return "Closed Caption Select";
            case 0x0063: return "VCR Or TV";
            case 0x0064: return "Broadcast Mode";
            case 0x0065: return "Snapshot";
            case 0x0066: return "Still";
            case 0x0080: return "Selection";
            case 0x0081: return "Assign";
            case 0x0082: return "Mode Step";
            case 0x0083: return "Recall Last";
            case 0x0084: return "Enter Channel";
            case 0x0085: return "Order Movie";
            case 0x0086: return "Channel";
            case 0x0087: return "Media Selection";
            case 0x0088: return "Media Select Computer";
            case 0x0089: return "Media Select TV";
            case 0x008A: return "Media Select WWW";
            case 0x008B: return "Media Select DVD";
            case 0x008E: return "Media Select Video Phone";
            case 0x008F: return "Media Select Games";
            case 0x0090: return "Media Select Messages";
            case 0x0091: return "Media Select CD";
            case 0x0092: return "Media Select VCR";
            case 0x0093: return "Media Select Tuner";
            case 0x0094: return "Quit";
            case 0x0095: return "Help";
            case 0x0096: return "Media Select Tape";
            case 0x0097: return "Media Select Cable";
            case 0x0098: return "Media Select Satellite";
            case 0x0099: return "Media Select Security";
            case 0x009A: return "Media Select Home";
            case 0x009B: return "Media Select Call";
            case 0x009C: return "Channel Increment";
            case 0x009D: return "Channel Decrement";
            case 0x009E: return "Media";
            case 0x00A0: return "VCR Plus";
            case 0x00A1: return "Once";
            case 0x00A2: return "Daily";
            case 0x00A3: return "Weekly";
            case 0x00A4: return "Monthly";
            case 0x00B0: return "Play";
            case 0x00B1: return "Pause";
            case 0x00B2: return "Record";
            case 0x00B3: return "Fast Forward";
            case 0x00B4: return "Rewind";
            case 0x00B5: return "Scan Next Track";
            case 0x00B6: return "Scan Previous Track";
            case 0x00B7: return "Stop";
            case 0x00B8: return "Eject";
            case 0x00B9: return "Random Play";
            case 0x00BA: return "Select Disc";
            case 0x00BB: return "Enter Disc";
            case 0x00BC: return "Repeat";
            case 0x00BD: return "Tracking";
            case 0x00BE: return "Track Normal";
            case 0x00BF: return "Slow Tracking";
            case 0x00C0: return "Frame Forward";
            case 0x00C1: return "Frame Back";
            case 0x00C2: return "Mark";
            case 0x00C3: return "Clear Mark";
            case 0x00C4: return "Repeat From Mark";
            case 0x00C5: return "Return To Mark";
            case 0x00C6: return "Search Mark Forward";
            case 0x00C7: return "Search Mark Backwards";
            case 0x00C8: return "Counter Reset";
            case 0x00C9: return "Show Counter";
            case 0x00CA: return "Tracking Increment";
            case 0x00CB: return "Tracking Decrement";
            case 0x00CC: return "Stop Or Eject";
            case 0x00CD: return "Play Or Pause";
            case 0x00CE: return "Play Or Skip";
            case 0x00E0: return "Volume";
            case 0x00E1: return "Balance";
            case 0x00E2: return "Mute";
            case 0x00E3: return "Bass";
            case 0x00E4: return "Treble";
            case 0x00E5: return "Bass Boost";
            case 0x00E6: return "Surround Mode";
            case 0x00E7: return "Loudness";
            case 0x00E8: return "MPX";
            case 0x00E9: return "Volume Increment";
            case 0x00EA: return "Volume Decrement";
            case 0x00F0: return "Speed";
            case 0x00F1: return "Playback Speed";
            case 0x00F2: return "Standard Play";
            case 0x00F3: return "Long Play";
            case 0x00F4: return "Extended Play";
            case 0x00F5: return "Slow";
            case 0x0100: return "Fan Enable";
            case 0x0101: return "Fan Speed";
            case 0x0102: return "Light Enable";
            case 0x0103: return "Light Illumination Level";
            case 0x0104: return "Climate Control Enable";
            case 0x0105: return "Room Temperature";
            case 0x0106: return "Security Enable";
            case 0x0107: return "Fire Alarm";
            case 0x0108: return "Police Alarm";
            case 0x0150: return "Balance Right";
            case 0x0151: return "Balance Left";
            case 0x0152: return "Bass Increment";
            case 0x0153: return "Bass Decrement";
            case 0x0154: return "Treble Increment";
            case 0x0155: return "Treble Decrement";
            case 0x0160: return "Speaker System";
            case 0x0161: return "Channel Left";
            case 0x0162: return "Channel Right";
            case 0x0164: return "Channel Front";
            case 0x0165: return "Channel Center Front";
            case 0x0166: return "Channel Side";
            case 0x0167: return "Channel Surround";
            case 0x0168: return "Channel Low Frequency Enhancement";
            case 0x0169: return "Channel Top";
            case 0x016A: return "Channel Unknown";
            case 0x0170: return "Sub Channel";
            case 0x0171: return "Sub Channel Increment";
            case 0x0172: return "Sub Channel Decrement";
            case 0x0173: return "Alternate Audio Increment";
            case 0x0174: return "Alternate Audio Decrement";
            case 0x0180: return "Application Launch Buttons";
            case 0x0181: return "AL Launch Button Configuration Tool";
            case 0x0182: return "AL Programmable Button Configuration";
            case 0x0183: return "AL Consumer Control Configuration";
            case 0x0184: return "AL Word Processor";
            case 0x0185: return "AL Text Editor";
            case 0x0186: return "AL Spreadsheet";
            case 0x0187: return "AL Graphics Editor";
            case 0x0188: return "AL Presentation App";
            case 0x0189: return "AL Database App";
            case 0x018A: return "AL Email Reader";
            case 0x018B: return "AL Newsreader";
            case 0x018C: return "AL Voicemail";
            case 0x018D: return "AL Contacts Or AddressBook";
            case 0x018E: return "AL Calendar Or Schedule";
            case 0x018F: return "AL Task Or Project Manager";
            case 0x0190: return "AL Log Or Journal Or Timecard";
            case 0x0191: return "AL Checkbook Or Finance";
            case 0x0192: return "AL Calculator";
            case 0x0193: return "AL A Or V Capture Or Playback";
            case 0x0194: return "AL Local Machine Browser";
            case 0x0195: return "AL LAN Or WAN Browser";
            case 0x0196: return "AL Internet Browser";
            case 0x0197: return "AL Remote Networking Or ISP Connect";
            case 0x0198: return "AL Network Conference";
            case 0x0199: return "AL Network Chat";
            case 0x019A: return "AL Telephony Or Dialer";
            case 0x019B: return "AL Logon";
            case 0x019C: return "AL Logoff";
            case 0x019D: return "AL Logon Or Logoff";
            case 0x019E: return "AL Terminal Lock Or Screensaver";
            case 0x019F: return "AL Control Panel";
            case 0x01A0: return "AL Command Line Processor Or Run";
            case 0x01A1: return "AL Process Or Task Manager";
            case 0x01A2: return "AL";
            case 0x01A3: return "AL Next Task Or Application";
            case 0x01A4: return "AL Previous Task Or Application";
            case 0x01A5: return "AL Preemptive Halt Task Or Application";
            case 0x0200: return "Generic GUI Application Controls";
            case 0x0201: return "AC New";
            case 0x0202: return "AC Open";
            case 0x0203: return "AC Close";
            case 0x0204: return "AC Exit";
            case 0x0205: return "AC Maximize";
            case 0x0206: return "AC Minimize";
            case 0x0207: return "AC Save";
            case 0x0208: return "AC Print";
            case 0x0209: return "AC Properties";
            case 0x021A: return "AC Undo";
            case 0x021B: return "AC Copy";
            case 0x021C: return "AC Cut";
            case 0x021D: return "AC Paste";
            case 0x021E: return "AC";
            case 0x021F: return "AC Find";
            case 0x0220: return "AC Find and Replace";
            case 0x0221: return "AC Search";
            case 0x0222: return "AC Go To";
            case 0x0223: return "AC Home";
            case 0x0224: return "AC Back";
            case 0x0225: return "AC Forward";
            case 0x0226: return "AC Stop";
            case 0x0227: return "AC Refresh";
            case 0x0228: return "AC Previous Link";
            case 0x0229: return "AC Next Link";
            case 0x022A: return "AC Bookmarks";
            case 0x022B: return "AC History";
            case 0x022C: return "AC Subscriptions";
            case 0x022D: return "AC Zoom In";
            case 0x022E: return "AC Zoom Out";
            case 0x022F: return "AC Zoom";
            case 0x0230: return "AC Full Screen View";
            case 0x0231: return "AC Normal View";
            case 0x0232: return "AC View Toggle";
            case 0x0233: return "AC Scroll Up";
            case 0x0234: return "AC Scroll Down";
            case 0x0235: return "AC Scroll";
            case 0x0236: return "AC Pan Left";
            case 0x0237: return "AC Pan Right";
            case 0x0238: return "AC Pan";
            case 0x0239: return "AC New Window";
            case 0x023A: return "AC Tile Horizontally";
            case 0x023B: return "AC Tile Vertically";
            case 0x023C: return "AC Format";
            default: return NULL;
        }
    }
    else if (page == kUsage_PagePID)
    {
        switch (usage)
        {
            case 1: return "Physical Interface Device";
            case 0x20: return "Normal";
            case 0x21: return "Set Effect Report";
            case 0x22: return "Effect Block Index";
            case 0x23: return "Parameter Block Offset";
            case 0x24: return "ROM Flag";
            case 0x25: return "Effect Type";
            case 0x26: return "ET Constant Force";
            case 0x27: return "ET Ramp";
            case 0x28: return "ET Custom Force Data";
            case 0x30: return "ET Square";
            case 0x31: return "ET Sine";
            case 0x32: return "ET Triangle";
            case 0x33: return "ET Sawtooth Up";
            case 0x34: return "ET Sawtooth Down";
            case 0x40: return "ET Spring";
            case 0x41: return "ET Damper";
            case 0x42: return "ET Inertia";
            case 0x43: return "ET Friction";
            case 0x50: return "Duration";
            case 0x51: return "Sample Period";
            case 0x52: return "Gain";
            case 0x53: return "Trigger Button";
            case 0x54: return "Trigger Repeat Interval";
            case 0x55: return "Axes Enable";
            case 0x56: return "Direction Enable";
            case 0x57: return "Direction";
            case 0x58: return "Type Specific Block Offset";
            case 0x59: return "Block Type";
            case 0x5a: return "Set Envelope Report";
            case 0x5b: return "Attack Level";
            case 0x5c: return "Attack Time";
            case 0x5d: return "Fade Level";
            case 0x5e: return "Fade Time";
            case 0x5f: return "Set Condition Report";
            case 0x60: return "CP Offset";
            case 0x61: return "Positive Coefficient";
            case 0x62: return "Negative Coefficient";
            case 0x63: return "Positive Saturation";
            case 0x64: return "Negative Saturation";
            case 0x65: return "Dead Band";
            case 0x66: return "Download Force Data Report";
            case 0x67: return "Isoch Custom Force Enable";
            case 0x68: return "Custom Force Data Report";
            case 0x69: return "Custom Force Data";
            case 0x6a: return "Custom Force Vendor Defined Data";
            case 0x6b: return "Set Custom Force Report";
            case 0x6c: return "Custom Force Data Offset";
            case 0x6d: return "Sample Count";
            case 0x6e: return "Set Periodic Report";
            case 0x6f: return "Offset";
            case 0x70: return "Magnitude";
            case 0x71: return "Phase";
            case 0x72: return "Period";
            case 0x73: return "Set Constant Force Report";
            case 0x74: return "Set Constant Force";
            case 0x75: return "Ramp Start";
            case 0x76: return "Ramp End";
            case 0x77: return "Effect Operation Report";
            case 0x78: return "Effect Operation";
            case 0x79: return "Op Effect Start";
            case 0x7a: return "Op Effect Start Solo";
            case 0x7b: return "Op Effect Stop";
            case 0x7c: return "Loop Count";
            case 0x7d: return "Gain Report";
            case 0x7e: return "Gain";
            case 0x7f: return "PID Pool Report";
            case 0x80: return "RAM Pool Size";
            case 0x81: return "ROM Pool Size";
            case 0x82: return "ROM Effect Block Count";
            case 0x83: return "Simultaneous Effects Max";
            case 0x84: return "Pool Alignment";
            case 0x85: return "PID Pool Move Report";
            case 0x86: return "Move Source";
            case 0x87: return "Move Destination";
            case 0x88: return "Move Length";
            case 0x89: return "PID Block Load Report";
            case 0x8b: return "Block Load Status";
            case 0x8c: return "Block Load Success";
            case 0x8d: return "Block Load Full";
            case 0x8e: return "Block Load Error";
            case 0x8f: return "Block Handle";
            case 0x90: return "PID Block Free Report";
            case 0x91: return "Type Specific Block Handle";
            case 0x92: return "PID State Report";
            case 0x94: return "Effect Playing";
            case 0x95: return "PID Device Control Report";
            case 0x96: return "PID Device Control";
            case 0x97: return "DC Enable Actuators";
            case 0x98: return "DC Disable Actuators";
            case 0x99: return "DC Stoop All Effects";
            case 0x9a: return "DC Device Reset";
            case 0x9b: return "DC Device Pause";
            case 0x9c: return "DC Device Continue";
            case 0x9f: return "Device Paused";
            case 0xa0: return "Actuators Enabled";
            case 0xa4: return "Safety Switch";
            case 0xa5: return "Actuator Override Switch";
            case 0xa6: return "Actuator Power";
            case 0xa7: return "Start Delay";
            case 0xa8: return "Parameter Block Size";
            case 0xa9: return "Device Managed Pool";
            case 0xaa: return "Shared parameter blocks";
            case 0xab: return "Create New Effect Report";
            case 0xac: return "RAM Pool Available";
                
                
            default: return NULL;
        }
    } 
    else if (page == kUsage_PagePowerDevice)
    {
        switch (usage)
        {
            case 0x0000: return "Undefined Usage";
            case 0x0001: return "Name Index";
            case 0x0002: return "Present Status";
            case 0x0003: return "Changed Status";
            case 0x0004: return "Uninterruptible Power Supply";
            case 0x0005: return "Power Supply";
            case 0x0010: return "Battery System power module";
            case 0x0011: return "Battery System ID";
            case 0x0012: return "Battery";
            case 0x0013: return "Battery ID";
            case 0x0014: return "Charger";
            case 0x0015: return "Charger ID";
            case 0x0016: return "Power Converter power module";
            case 0x0017: return "Power Converter ID";
            case 0x0018: return "Outlet System power module";
            case 0x0019: return "Outlet System ID";
            case 0x001A: return "Input";
            case 0x001B: return "Input ID";
            case 0x001C: return "Output";
            case 0x001D: return "Output ID";
            case 0x001E: return "Flow";
            case 0x001F: return "Flow ID";
            case 0x0020: return "Outlet";
            case 0x0021: return "Outlet ID";
            case 0x0022: return "Gang";
            case 0x0023: return "Gang ID";
            case 0x0024: return "Power Summary";
            case 0x0025: return "Power Summary ID";
            case 0x0030: return "Voltage";
            case 0x0031: return "Current";
            case 0x0032: return "Frequency";
            case 0x0033: return "Apparent Power";
            case 0x0034: return "RMS Power";
            case 0x0035: return "Percent Load";
            case 0x0036: return "Temperature";
            case 0x0037: return "Humidity";
            case 0x0038: return "Bad Condition Count";
            case 0x0040: return "Nominal Voltage";
            case 0x0041: return "Nominal Current";
            case 0x0042: return "Nominal Frequency";
            case 0x0043: return "Nominal Apparent Power";
            case 0x0044: return "Nominal RMS Power";
            case 0x0045: return "Nominal Percent Load";
            case 0x0046: return "Nominal Temperature";
            case 0x0047: return "Nominal Humidity";
            case 0x0050: return "Switch On Control";
            case 0x0051: return "Switch Off Control";
            case 0x0052: return "Toogle Sequence Control";
            case 0x0053: return "Min Transfer Voltage";
            case 0x0054: return "Max Transfer Voltage";
            case 0x0055: return "Delay Before Reboot";
            case 0x0056: return "Delay Before Startup";
            case 0x0057: return "Delay Before Shutdown";
            case 0x0058: return "Test Request/Result";
            case 0x0059: return "Reset Request/Result";
            case 0x005A: return "Audible Alarm Control";
            case 0x0060: return "Present";
            case 0x0061: return "Good";
            case 0x0062: return "Internal Failure";
            case 0x0063: return "Voltage Out Of Range";
            case 0x0064: return "Frequency Out Of Range";
            case 0x0065: return "Overload";
            case 0x0066: return "Over Charged";
            case 0x0067: return "Over Temperature";
            case 0x0068: return "Shutdown Requested";
            case 0x0069: return "Shutdown Imminent";
            case 0x006B: return "On/Off Switch Status";
            case 0x006C: return "Switchable";
            case 0x006D: return "Used";
            case 0x006E: return "Boosted";
            case 0x006F: return "Bucked";
            case 0x0070: return "Initialized";
            case 0x0071: return "Tested";
            case 0x0072: return "Awaiting Power";
            case 0x0073: return "Communication Lost";
            case 0x00FD: return "Manufacturer String Index";
            case 0x00FE: return "Product String Index";
            case 0x00FF: return "Serial Number String Index";
            default: return NULL;
        }
    }
    else if (page == kUsage_PageBatterySystem)
    {
        switch (usage)
        {
            case 0x0000: return "Undefined Usage";
            case 0x0001: return "Mode";
            case 0x0002: return "Status";
            case 0x0003: return "Alarm Warning";
            case 0x0004: return "Charger Mode";
            case 0x0005: return "Charger Status";
            case 0x0006: return "Charger Extended Status";
            case 0x0007: return "Selector State";
            case 0x0008: return "Selector Presets";
            case 0x0009: return "Selector Info";
            case 0x0010: return "Mfg Function 1";
            case 0x0011: return "Mfg Function 2";
            case 0x0012: return "Mfg Function 3";
            case 0x0013: return "Mfg Function 4";
            case 0x0014: return "Mfg Function 5";
            case 0x0015: return "Connection To System Management Bus";
            case 0x0016: return "Output Connection Status";
            case 0x0017: return "Charger Connection";
            case 0x0018: return "Battery Insertion";
            case 0x0019: return "Use Next";
            case 0x001A: return "OK To Use";
            case 0x001B: return "Battery Supported";
            case 0x001C: return "Selector Revision";
            case 0x001D: return "Charging Indicator";
            case 0x0028: return "Manufacturer Access";
            case 0x0029: return "Remaining Capacity Limit";
            case 0x002A: return "Remaining Time Limit";
            case 0x002B: return "At Rate...";
            case 0x002C: return "Capacity Mode";
            case 0x002D: return "Broadcast To Charger";
            case 0x002E: return "Primary Battery";
            case 0x002F: return "Charge Controller";
            case 0x0040: return "Terminate Charge";
            case 0x0041: return "Terminate Discharge";
            case 0x0042: return "Below Remaining Capacity Limit";
            case 0x0043: return "Remaining Time Limit Expired";
            case 0x0044: return "Charging";
            case 0x0045: return "Discharging";
            case 0x0046: return "Fully Charged";
            case 0x0047: return "Fully Discharged";
            case 0x0048: return "Conditioning Flag";
            case 0x0049: return "At Rate OK";
            case 0x004A: return "SMB Error Code";
            case 0x004B: return "Need Replacement";
            case 0x0060: return "At Rate Time To Full";
            case 0x0061: return "At Rate Time To Empty";
            case 0x0062: return "Average Current";
            case 0x0063: return "Max Error";
            case 0x0065: return "Absolute State Of Charge";
            case 0x0066: return "Remaining Capacity";
            case 0x0067: return "Full Charge Capacity";
            case 0x0068: return "Run Time To Empty";
            case 0x0069: return "Average Time To Empty";
            case 0x006A: return "Average Time To Full";
            case 0x006B: return "Cycle Count";
            case 0x0080: return "Batt Pack Model Level";
            case 0x0081: return "Internal Charge Controller";
            case 0x0082: return "Primary Battery Support";
            case 0x0083: return "Design Capacity";
            case 0x0084: return "Specification Info";
            case 0x0085: return "Manufacturer Date";
            case 0x0086: return "Serial Number";
            case 0x0087: return "Manufacturer Name Index";
            case 0x0088: return "Device Name Index";
            case 0x0089: return "Device Chemistry Index";
            case 0x008A: return "Manufacturer Data";
            case 0x008B: return "Rechargable";
            case 0x008C: return "Warning Capacity Limit";
            case 0x008D: return "Capacity Granularity 1";
            case 0x008E: return "Capacity Granularity 2";
            case 0x008F: return "OEM Information Index";
            case 0x00C0: return "Inhibit Charge";
            case 0x00C1: return "Enable Polling";
            case 0x00C2: return "Reset To Zero";
            case 0x00D0: return "AC Present";
            case 0x00D1: return "Battery Present";
            case 0x00D2: return "Power Fail";
            case 0x00D3: return "Alarm Inhibited";
            case 0x00D4: return "Thermistor Under Range";
            case 0x00D5: return "Thermistor Hot";
            case 0x00D6: return "Thermistor Cold";
            case 0x00D7: return "Thermistor Over Range";
            case 0x00D8: return "Voltage Out Of Range";
            case 0x00D9: return "Current Out Of Range";
            case 0x00DA: return "Current Not Regulated";
            case 0x00DB: return "Voltage Not Regulated";
            case 0x00DC: return "Master Mode";
            case 0x00F0: return "Charger Support Selector";
            case 0x00F1: return "Charger Specification";
            case 0x00F2: return "Charger Level 2";
            case 0x00F3: return "Charger Level 3";
            default: return NULL;
        }
    } 
    
    return NULL;
}
