/*
 *
 * Copyright 2012 Philip Quinn. All Rights Reserved.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */
//
//  main.c
//  HIDReportTracer
//

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdint.h>
#include <getopt.h>
#include <sys/time.h>
#include <mach/mach_time.h>
#include <CoreServices/CoreServices.h>
#include <CoreFoundation/CoreFoundation.h>
#include <IOKit/hid/IOHIDManager.h>
#include "HIDDescriptors.h"

#pragma mark Delicious Globals

static struct {
    // Time the process started (in mach_absolute_time() units)
    uint64_t startTime;       
    uint32_t usagePageFilter;
    uint32_t usageFilter;
    uint32_t vendorFilter;
    uint32_t productFilter;
    uint8_t raw:1;
} globalArgs = {0};

// Dictionary of {IOHIDElement: CFNumber}, where the CFNumber contains a uint64_t
// (as an kCFNumberSInt64Type) representing the timestamp of the last change to
// the element.
static CFMutableDictionaryRef gLastElementTimes = NULL;

#pragma mark Static Methods

/**
 * Prints a CFString to a file.
 */
static void fprint_CFString(FILE *file, CFStringRef str) {
    if (!str) return;
    
    CFIndex len = CFStringGetLength(str);
    CFStringInlineBuffer buffer;
    
    CFStringInitInlineBuffer(str, &buffer, CFRangeMake(0, len));
    for (CFIndex idx = 0; idx < len; idx++) {
        UniChar ch = CFStringGetCharacterFromInlineBuffer(&buffer, idx);
        // From CFUtilities.c
        if (ch < 128) {
            fprintf(file, "%c", ch);
        } else {
            fprintf(file, "\\u%04x", ch);
        }
    }
}

/**
 * Converts an OS absolute time (period in the units of mach_absolute_time())
 * into milliseconds.
 * @see https://developer.apple.com/library/mac/#qa/qa1398/_index.html
 */
static inline uint64_t convert_absolute_time(uint64_t absolute) {
    Nanoseconds nanos = AbsoluteToNanoseconds(*(AbsoluteTime*)&absolute);
    return (*(uint64_t*)&nanos)/1000000;
}

/**
 * Prints out the human-readable names of the given usage page and usage
 * (if available).
 */
static void fprint_usage(FILE* file, uint32_t page, uint32_t usage) {
    const char* pagestr = usage_page_name(page);
    const char* usagestr = usage_name(page, usage);
    
    if (globalArgs.raw || (!pagestr && !usagestr))
        fprintf(file, "0x%02X/0x%02X", page, usage);
    else if (pagestr && !usagestr)
        fprintf(file, "%s/0x%02X", pagestr, usage);
    else if (!pagestr && usagestr)
        fprintf(file, "0x%02X/%s", page, usagestr);
    else 
        fprintf(file, "%s/%s", pagestr, usagestr);
}

#pragma mark Callbacks

/**
 * Callback when a HID's report changes an input value.
 */
static void value_changed(void *context, IOReturn result, void* sender, IOHIDValueRef value) {
    //IOHIDDeviceRef device = (IOHIDDeviceRef)sender;
    IOHIDElementRef element = IOHIDValueGetElement(value);
    uint64_t timestamp = IOHIDValueGetTimeStamp(value);
    
    // Optional filter on usage/page
    uint32_t page = IOHIDElementGetUsagePage(element);
    if (globalArgs.usagePageFilter != 0 && page != globalArgs.usagePageFilter)
        return;
    uint32_t usage = IOHIDElementGetUsage(element);
    if (globalArgs.usageFilter != 0 && usage != globalArgs.usageFilter)
        return;
    
    // See if this is an actual update for the value, or is an update for another value
    // that happens to be in the same report.
    if (CFDictionaryContainsKey(gLastElementTimes, element)) {
        CFNumberRef lastUpdate = (CFNumberRef)CFDictionaryGetValue(gLastElementTimes, element);
        uint64_t t_last = 0;
        CFNumberGetValue(lastUpdate, kCFNumberSInt64Type, &t_last);
        if (t_last && t_last == timestamp)
            return;
    }
    CFNumberRef update = CFNumberCreate(NULL, kCFNumberSInt64Type, &timestamp);
    CFDictionarySetValue(gLastElementTimes, element, update);
    CFRelease(update);

    // Print timestamp
    // IOHIDValueGetTimestamp() returns an OS timestamp; convert it into 
    // milliseconds since this program started.
    uint64_t t_off = timestamp - globalArgs.startTime;
    printf("%llu", convert_absolute_time(t_off));
    //fprint_CFAbsoluteTime(stdout, CFAbsoluteTimeGetCurrent());
    printf("\t");

    // Print element usage page/usage id
    fprint_usage(stdout, page, usage);
    printf("\t");
    
    // If the length fits into a nice integer, print it as an integer;
    // otherwise, print the raw bytes.
    if (!globalArgs.raw && IOHIDValueGetLength(value) <= sizeof(CFIndex)) {
        //printf("%ld (calibrated: %.2f, physical: %.2f)\n",
        //        IOHIDValueGetIntegerValue(value),
        //        IOHIDValueGetScaledValue(value, kIOHIDValueScaleTypeCalibrated),
        //        IOHIDValueGetScaledValue(value, kIOHIDValueScaleTypePhysical));
        printf("%ld\n", IOHIDValueGetIntegerValue(value));
    } else {
        const uint8_t *bytes = IOHIDValueGetBytePtr(value);
        CFIndex len = IOHIDValueGetLength(value);
        for (CFIndex i = 0; i < len; i++)
            printf("%02X ", bytes[i]);
        printf("\n");
    }
}

/**
 * Callback when a device matching the criteria is connected.
 */
static void device_found(void *context, IOReturn result, void* sender, IOHIDDeviceRef device) {
    CFStringRef manufacturer = (CFStringRef)IOHIDDeviceGetProperty(device, CFSTR(kIOHIDManufacturerKey));
    CFStringRef product = (CFStringRef)IOHIDDeviceGetProperty(device, CFSTR(kIOHIDProductKey));
    printf("%lld", convert_absolute_time(mach_absolute_time() - globalArgs.startTime));
    printf(" ");
    printf("Device connected: ");
    fprint_CFString(stdout, product);
    printf(" (");
    fprint_CFString(stdout, manufacturer);
    printf(")\n");
    
    // Print its registry path
    io_object_t io_dev = IOHIDDeviceGetService(device);
    io_string_t io_path;
    kern_return_t err = IORegistryEntryGetPath(io_dev, kIOServicePlane, io_path);
    if (err == KERN_SUCCESS) {
        printf("\t%s\n", io_path);
    }
    
    // Print its inputs
    CFArrayRef elements = IOHIDDeviceCopyMatchingElements(device, NULL, kIOHIDOptionsTypeNone);
    CFIndex len = CFArrayGetCount(elements);
    for (CFIndex i = 0; i < len; i++) {
        IOHIDElementRef element = (IOHIDElementRef)CFArrayGetValueAtIndex(elements, i);
        
        // Ignore outputs, features, and collections, as they won't be reported
        if (IOHIDElementGetType(element) > kIOHIDElementTypeInput_ScanCodes) {
            continue;
        }
        
        uint32_t page = IOHIDElementGetUsagePage(element);
        uint32_t usage = IOHIDElementGetUsage(element);
        
        printf("\t");
        fprint_usage(stdout, page, usage);
        
        uint32_t reportSize = IOHIDElementGetReportSize(element);
        uint32_t reportCount = IOHIDElementGetReportCount(element);
        printf("\t%dx%d\n", reportSize, reportCount);
    }
    CFRelease(elements);
    
    // Report input value callback
    IOHIDDeviceRegisterInputValueCallback(device, value_changed, NULL);
}

/**
 * Callback when a device matching the criteria is disconnected.
 */
static void device_lost(void *context, IOReturn result, void* sender, IOHIDDeviceRef device) {
    CFStringRef manufacturer = (CFStringRef)IOHIDDeviceGetProperty(device, CFSTR(kIOHIDManufacturerKey));
    CFStringRef product = (CFStringRef)IOHIDDeviceGetProperty(device, CFSTR(kIOHIDProductKey));
    printf("%lld", convert_absolute_time(mach_absolute_time() - globalArgs.startTime));
    printf(" ");
    printf("Device disconnected: ");
    fprint_CFString(stdout, product);
    printf(" (");
    fprint_CFString(stdout, manufacturer);
    printf(")\n");
}

/**
 * Sets up a IOHIDManager to watch for devices matching the given vendor and product IDs.
 * Returns an IOHIDManagerRef that must be scheduled (and later released) by the caller.
 */
IOHIDManagerRef setup_trace(int vendorID, int productID) {
    IOHIDManagerRef manager = IOHIDManagerCreate(NULL, kIOHIDOptionsTypeNone);
    assert(manager != NULL);
    
    // Only look for devices with matching vendor/product IDs
    CFStringRef keys[2] = {CFSTR(kIOHIDVendorIDKey), CFSTR(kIOHIDProductIDKey)};
    CFNumberRef vals[2] = {CFNumberCreate(NULL, kCFNumberIntType, &vendorID),
                           CFNumberCreate(NULL, kCFNumberIntType, &productID)};
    
    CFDictionaryRef criteria = CFDictionaryCreate(NULL, (const void**)keys, (const void**)vals, 2, 
                                                  &kCFTypeDictionaryKeyCallBacks, 
                                                  &kCFTypeDictionaryValueCallBacks);
    IOHIDManagerSetDeviceMatching(manager, criteria);
    CFRelease(criteria);
    CFRelease(vals[0]);
    CFRelease(vals[1]);
    
    // Register callbacks
    IOHIDManagerRegisterDeviceMatchingCallback(manager, device_found, NULL);
    IOHIDManagerRegisterDeviceRemovalCallback(manager, device_lost, NULL);
    
    return manager;
}

#pragma mark Main

void usage() {
        fprintf(stderr,
                "Usage: HIDReportTracer -v vid -d pid [-p page] [-u usage] [-r]\n"
                "\n"
                "OPTIONS\n"
                "    -v vid\n"
                "    --vendor vid\n"
                "        The vendor ID of the device to inspect.\n"
                "    -d pid\n"
                "    --device pid\n"
                "        The product ID of the device to inspect.\n"
                "    -p page\n"
                "    --page page\n"
                "        Filter inputs to show only those for the given usage page.\n"
                "    -u usage\n"
                "    --usage usage\n"
                "        Filter inputs to show only those for the given usage.\n"
                "    -r\n"
                "    --raw\n"
                "        Turn off value formatting.\n"
                "\n");
}

int main(int argc, char * const argv[])
{
    const char *optString = ":v:d:p:u:r";
    const struct option longOpts[] = {
        {"vendor", required_argument, NULL, 'v'},
        {"device", required_argument, NULL, 'd'},
        {"page", required_argument, NULL, 'p'},
        {"usage", required_argument, NULL, 'u'},
        {"raw", no_argument, NULL, 'r'},
        {NULL, no_argument, NULL, 0}
    };
    
    // Parse options
    char c;
    while ((c = getopt_long(argc, argv, optString, longOpts, NULL)) != -1) {
        switch(c) {
            case 'v':
                globalArgs.vendorFilter = (uint32_t)strtol(optarg, NULL, 0);
                break;
            case 'd':
                globalArgs.productFilter = (uint32_t)strtol(optarg, NULL, 0);
                break;
            case 'p':
                globalArgs.usagePageFilter  = (uint32_t)strtol(optarg, NULL, 0);
                break;
            case 'u':
                globalArgs.usageFilter  = (uint32_t)strtol(optarg, NULL, 0);
                break;
            case 'r':
                globalArgs.raw = 1;
                break;
            case ':':
                fprintf(stderr, "Option -%c requires an operand\n", optopt);
                usage();
                exit(1);
            case '?':
                fprintf(stderr, "Unrecognised option: -%c\n", optopt);
                usage();
                exit(1);
            default:
                usage();
                exit(1);
        }
    }
    if (globalArgs.vendorFilter == 0 || globalArgs.productFilter == 0) {
        usage();
        exit(1);
    }
    
    gLastElementTimes = CFDictionaryCreateMutable(NULL, 10, 
                                                  &kCFTypeDictionaryKeyCallBacks,
                                                  &kCFTypeDictionaryValueCallBacks);
    globalArgs.startTime = mach_absolute_time();
    
    // Setup trace
    IOHIDManagerRef manager = setup_trace(globalArgs.vendorFilter, globalArgs.productFilter);
    IOHIDManagerScheduleWithRunLoop(manager, CFRunLoopGetCurrent(), kCFRunLoopDefaultMode);
    
    // Start trace
    IOReturn err = IOHIDManagerOpen(manager, kIOHIDOptionsTypeNone);
    if (err != kIOReturnSuccess) {
        CFErrorRef error = CFErrorCreate(NULL, kCFErrorDomainOSStatus, err, NULL);
        CFStringRef errorDescription = CFErrorCopyDescription(error);
        fprintf(stderr, "Error opening IOHIDManager: ");
        fprint_CFString(stderr, errorDescription);
        fprintf(stderr, " (%ld)\n", CFErrorGetCode(error));
        CFRelease(errorDescription);
        CFRelease(error);
        CFRelease(manager);
        exit(1);
    }
    
    // Never returns...
    CFRunLoopRun();
    
    // End trace
    CFRelease(manager);
    CFRelease(gLastElementTimes);
    
    return 0;
}

