/*
 *
 * Copyright 2012-2013 Philip Quinn. All Rights Reserved.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */
//
//  activities.hpp
//  Scroll Gain Inspector
//

#ifndef ACTIVITIES_HPP
#define ACTIVITIES_HPP

#include <functional>
#include <chrono>
#include <memory>
#include <mutex>
#include <thread>
#include <queue>
#include <stdint.h>
#include "devices.hpp"

using std::chrono::system_clock;
using std::chrono::milliseconds;

/**
 * Issues scrolling events onto a ScrollingDevice.
 */
class ScrollingActivity {
private:
    // Interval between events
    milliseconds _interval;
    // Magnitude of events
    int8_t _magnitude;
    // Mutex that is unlocked when the thread should stop
    std::mutex _stop;
    std::unique_ptr<std::thread> _dispatcher;
    // Device to send events to
    ScrollingDevice &_device;
    
public:
    ScrollingActivity(ScrollingDevice &dev) 
        : _interval(milliseconds(0)), _magnitude(-1), _device(dev) {}
    
    ~ScrollingActivity() {
        this->stop();
    }
    
    void start() {
        if (_dispatcher)
            this->stop();
        _stop.lock();
        _dispatcher = std::unique_ptr<std::thread>(new std::thread(std::ref(*this)));
    }
    
    void stop() {
        if (_dispatcher && _dispatcher->joinable()) {
            _stop.unlock();
            _dispatcher->join();
        }
    }
    
    // Run on a separate thread
    void operator()() {
        this->reset();
        while(true) {
            if (_stop.try_lock()) {
                _stop.unlock();
                break;
            }
            
            this->scroll();
            if (this->interval().count() == 0)
                break;
            std::this_thread::sleep_for(this->interval());
        }
    }
    
    int8_t magnitude() const { return _magnitude; }
    milliseconds interval() const { return _interval; }
protected:
    virtual void reset() {}
    // Called repeatedly to send an event to the device and update the interval
    virtual void scroll() { this->device().scroll(_magnitude); }
    // Getters/setters for that
    ScrollingDevice& device() const { return _device; }
    void setInterval(milliseconds interval) { _interval = interval; }
    void setMagnitude(int8_t magnitude) { _magnitude = magnitude; }
};

/**
 * Issues scrolling events with a fixed magnitude and interval.
 * The minimum interval of the device is not respected.
 */
class ConstantScrolling : public ScrollingActivity {
public:
    /**
     * An interval of 0 will result in only one scroll event of magnitude m will be sent.
     */
    ConstantScrolling(ScrollingDevice &dev, int8_t m, std::chrono::milliseconds d) 
        : ScrollingActivity(dev)  {
        setMagnitude(m);
        setInterval(d);
    }
};

/**
 * Issues scrolling events for a fixed velocity.
 * The magnitude and interval of these events are calcuated with respect to the
 * device's minimum interval.
 */
class ConstantVelocityScrolling : public ConstantScrolling {
private:
    std::chrono::time_point<system_clock> _startScroll;
    std::chrono::time_point<system_clock> _lastScroll;
    float _carry;
    float _targetVelocity;
public:
    // Velocity in counts per second
    ConstantVelocityScrolling(ScrollingDevice &dev, int velocity) 
        : ConstantScrolling(dev, 1, dev.minReportRate()) {

        int8_t sign = (velocity > 0 ? 1 : -1);
        // Try to configure the velocity such that magnitude = 1 until the delta
        // gets too small, then start increasing the magnitude.

        // Magnitude if reporting at the device's min rate
        double magnitude = abs(velocity)/(1000.f/dev.minReportRate().count());
        if (magnitude < 1.f) {
            setInterval(milliseconds(1000/abs(velocity)));
            setMagnitude(1*sign);
        } else {
            setInterval(dev.minReportRate());
            setMagnitude((int8_t)(magnitude*sign));
        }
    }
protected:
    virtual void reset() {
        ConstantScrolling::reset();
        _startScroll = system_clock::now();
        _lastScroll = _startScroll;
        _carry = 0.f;
        _targetVelocity = (1/(float)interval().count())*magnitude(); // counts per ms
    }

    virtual void scroll() {
        auto now = system_clock::now();
        if (_startScroll == _lastScroll) {
            _lastScroll = now;
            ConstantScrolling::scroll();
            return;
        }

        /* 
         * Since it's rare that the sleep() will wake up *exactly* when we want to,
         * calculate how much we should have travelled (given the target velocity)
         * since the event.
         */
        auto delta = std::chrono::duration_cast<milliseconds>(system_clock::now() - _lastScroll).count();
        float travel = (_targetVelocity * delta) + _carry;
        
        int8_t oldMagnitude = magnitude();
        setMagnitude((int8_t)travel);
        ConstantScrolling::scroll();
        setMagnitude(oldMagnitude);

        _carry = travel - (int8_t)travel;
        _lastScroll = now;
    }
};


class SteppedVelocityScrolling : public ConstantVelocityScrolling {
public:
    // Velocity in counts per second
    SteppedVelocityScrolling(ScrollingDevice &dev, int velocity) 
        : ConstantVelocityScrolling(dev, velocity) {
    }

    void faster() {
        auto i = interval();
        auto m = magnitude();
        
        if (i > device().minReportRate()) {
            // Drop interval at differing rates to adjust for differing resolution            
            if (i > milliseconds(500)) 
                setInterval(milliseconds(500));
            else if (i > milliseconds(300)) 
                setInterval(i - milliseconds(100));
            else if (i > milliseconds(100))
                setInterval(i - milliseconds(50));
            else if (i > milliseconds(80))
                setInterval(i - milliseconds(10));
            else if (i > milliseconds(50))
                setInterval(i - milliseconds(5));
            else if (i > milliseconds(35))
                setInterval(i - milliseconds(2));
            else
                setInterval(i - milliseconds(1));
        } else {
            
            if (m == INT8_MAX || m == INT8_MIN)
                return;
            
            // Account for direction sign
            if (m < 0)
                setMagnitude(m - 1);
            else
                setMagnitude(m + 1);
        }
    }
};

class SlidingVelocityScrolling : public SteppedVelocityScrolling {
public:
    SlidingVelocityScrolling(ScrollingDevice &dev, int velocity)
    : SteppedVelocityScrolling(dev, velocity) {}
    
    virtual void scroll() {
        SteppedVelocityScrolling::scroll();
        this->faster();
    }
};

/**
 * Issues a repeating set of scroll commands at a fixed magnitude.
 */
class QueueScrolling : public ScrollingActivity {
private:
    std::queue<milliseconds> _repeatQueue;
public: 
    QueueScrolling(ScrollingDevice &dev, int8_t magnitude=-1) 
        : ScrollingActivity(dev) {
        setMagnitude(magnitude);
    }
    
protected:
    virtual void scroll() {
        ScrollingActivity::scroll();
        
        auto nextInterval = _repeatQueue.front();
        _repeatQueue.pop();
        _repeatQueue.push(nextInterval);
        setInterval(nextInterval);
    }
    
    virtual void pushInterval(milliseconds ms) {
        _repeatQueue.push(ms);
    }
};

/**
 * Issues a repeating set of commands that simulates clutching.
 */
class ClutchingActivity : public QueueScrolling {
public:
    ClutchingActivity(ScrollingDevice &dev,
                      milliseconds clutchGap,
                      milliseconds longGap,
                      milliseconds medGap,
                      milliseconds shortGap,
                      int8_t magnitude=-1) : QueueScrolling(dev, magnitude) {
        this->pushInterval(clutchGap);
        this->pushInterval(longGap);
        this->pushInterval(shortGap);
        this->pushInterval(shortGap);
        this->pushInterval(shortGap);
        this->pushInterval(shortGap);
        this->pushInterval(medGap);
    }
};

#endif // ACTIVITIES_HPP
