/*
 *
 * Copyright 2012-2013 Philip Quinn. All Rights Reserved.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */
//
//  devices.hpp
//  Scroll Gain Inspector
//

#ifndef DEVICES_HPP
#define DEVICES_HPP

#include <cstdint>
#include <chrono>
#include "hidapi.h"

using std::chrono::milliseconds;

class ScrollingDevice {
private:
    hid_device *_device;
    unsigned short _vendorId;
    unsigned short _productId;
    // Shortest interval between events
    milliseconds _minReportRate;
    
public:
    ScrollingDevice(unsigned short vendorId, unsigned short productId,
                    milliseconds minReportRate=milliseconds(1)) 
        : _device(nullptr), _vendorId(vendorId), _productId(productId), _minReportRate(minReportRate) {}
    
    ~ScrollingDevice() {
        if (_device)
            this->close();
    }

    bool open() {
        if (_device) return true;
        _device = hid_open(_vendorId, _productId, L"EchoMouse");
        return _device ? true : false;
    }

    void close() {
        hid_close(_device);
        _device = nullptr;
    }
    
    milliseconds minReportRate() const { return _minReportRate; }

    bool scroll(int8_t amount) const {
#ifndef _WIN32
        unsigned char data[4] = {0, 0, 0, static_cast<unsigned char>(amount)};
        int res = hid_write(_device, data, sizeof(data));
#else
        // Windows opens certain HID devices for exclusive read/write; send
        // feature reports instead.
        // http://msdn.microsoft.com/en-us/library/ff543477.aspx
        unsigned char data[5] = {0, 0, 0, static_cast<unsigned char>(amount), 0};
        int res = hid_send_feature_report(_device, data, sizeof(data));
#endif
        return res > 0 ? true : false;
    }
};

class EchoMouse : public ScrollingDevice {
public:
    EchoMouse() : ScrollingDevice(0x4d8, 0x3f) {}
};

class MicrosoftDevice : public ScrollingDevice {
public:
    MicrosoftDevice(unsigned short productId, 
                    milliseconds minReportRate) 
        : ScrollingDevice(0x45e, productId, minReportRate) {}
};

class MicrosoftWheel : public MicrosoftDevice {
public:
    MicrosoftWheel() : MicrosoftDevice(0x40, milliseconds(8)) {}
};

class LogitechDevice : public ScrollingDevice {
public:
    LogitechDevice(unsigned short productId, 
                   milliseconds minReportRate) 
        : ScrollingDevice(0x46d, productId, minReportRate) {}
};

class LogitechMX518 : public LogitechDevice {
public:
    LogitechMX518() : LogitechDevice(0xc01e, milliseconds(10)) {}
};

class LogitechMX500 : public LogitechDevice {
public:
    LogitechMX500() : LogitechDevice(0xc025, milliseconds(10)) {}
};

class LogitechM500 : public LogitechDevice {
public:
    LogitechM500() : LogitechDevice(0xc069, milliseconds(8)) {}
};

class IBMScrollPoint : public ScrollingDevice {
public:
    IBMScrollPoint() : ScrollingDevice(0x4b3, 0x3108, milliseconds(10)) {}
};

#endif // DEVICES_HPP

