/*
 *
 * Copyright 2012-2013 Philip Quinn. All Rights Reserved.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */
//
//  experimental.hpp
//  Scroll Gain Inspector
//

#ifndef EXPERIMENTAL_HPP
#define EXPERIMENTAL_HPP

#include <cstdlib>
#include <list>
#include <numeric>
#include <iostream>
#include <iomanip>
#include <thread>
#include "devices.hpp"
#include "activities.hpp"

using std::chrono::time_point;
using std::chrono::seconds;
using std::chrono::milliseconds;
using std::chrono::system_clock;
using std::chrono::duration_cast;

/**
 * Runs an inspection experiment where events are issues to a ScrollingDevice
 * (via. a ScrollingActivity) and then system (UI) scroll events are read back
 * and printed.
 */
class ExperimentalScroller {
private:
    ScrollingDevice &_device;
    float _scrolledAmount;
    std::unique_ptr<std::thread> _dispatcher;
    
public:
    ExperimentalScroller(ScrollingDevice &dev) : _device(dev), _scrolledAmount(0.f) {}

    virtual void recordScroll(float amount) { _scrolledAmount += amount;  }
    virtual void reset() { _scrolledAmount = 0.f; }
    
    void run() {
        _dispatcher = std::unique_ptr<std::thread>(new std::thread(std::ref(*this)));
    }
    
    // Run on a separate thread
    virtual void operator()() {}
    
protected:
    ScrollingDevice& device() const { return _device; }
    float scrolledAmount() { return _scrolledAmount; }
};

class VelocityScroller : public ExperimentalScroller {
private:
    std::unique_ptr<SteppedVelocityScrolling> _activity;
    int _maxVelocity;
    int _startVelocity;
    seconds _sampleDuration;
    
public:
    VelocityScroller(ScrollingDevice &dev, int start, int max,
                     seconds sampleDuration=seconds(5)) 
        : ExperimentalScroller(dev),
          _maxVelocity(max), _startVelocity(start), _sampleDuration(sampleDuration) {}
    
    void operator()() {
        float actualVelocity = (float)_startVelocity;
        _activity = std::unique_ptr<SteppedVelocityScrolling>(new SteppedVelocityScrolling(this->device(), _startVelocity));
        this->reset();
        
        while ((_startVelocity > 0 && actualVelocity < _maxVelocity) ||
               (_startVelocity < 0 && actualVelocity > _maxVelocity)) {
            this->reset();
            _activity->start();
            std::this_thread::sleep_for(_sampleDuration);
            _activity->stop();

            float intervals = 1000.f / _activity->interval().count();
            actualVelocity = (intervals*_activity->magnitude());
            
            std::cout.setf(std::ios::fixed, std::ios::floatfield);
            std::cout << std::setprecision(2) <<  fabs(actualVelocity) << '\t' << 
                fabs(scrolledAmount()/_sampleDuration.count()) << std::endl;

            _activity->faster();            
            std::this_thread::sleep_for(seconds(2));
        }
    }
};

class ExperimentalVelocityScroller : public ExperimentalScroller {
private:
    time_point<system_clock> _startScroll;
    time_point<system_clock> _lastScroll;
    float _carry;
    bool _smoothing;
    std::list<float> _window;
    
public:
    ExperimentalVelocityScroller(ScrollingDevice &dev) : ExperimentalScroller(dev), _smoothing(true) {}
    
    virtual void reset() {
        ExperimentalScroller::reset();
        _startScroll = system_clock::now();
        _lastScroll = _startScroll;
        _window.clear();
        // If the delta is too small to record a velocity, carry the amount
        // onto the next event
        _carry = 0.f;
    }

    virtual void recordScroll(float amount) {
        amount += _carry;
        _carry = 0.f;
        
        auto now = system_clock::now();
        if (_startScroll == _lastScroll) {
            _lastScroll = now;
            return;
        }
       
        auto delta = now - _lastScroll;
        auto millis = duration_cast<milliseconds>(delta).count();
        if (millis == 0) {
            _carry += amount;
            return;
        }
        
        auto velocity = fabs(amount/millis);
        if (_smoothing) {
            _window.push_front(velocity);
            if (_window.size() > 3)
                _window.pop_back();
            velocity = std::accumulate(std::begin(_window), std::end(_window), 0.f) / _window.size();
        }
        
        auto offset = duration_cast<milliseconds>(now-_startScroll).count();        
        std::cout << offset << '\t' << velocity << std::endl;
        
        _lastScroll = now;
    }
};

class MaintainVelocityScroller : public ExperimentalVelocityScroller {
private:
    std::unique_ptr<ScrollingActivity> _activity;
    int _velocity;
public:
    MaintainVelocityScroller(ScrollingDevice &dev, int velocity) : ExperimentalVelocityScroller
        (dev), _velocity(velocity) {}

    void operator()() {
        _activity = std::unique_ptr<ScrollingActivity>(new ConstantVelocityScrolling(this->device(), _velocity));
        this->reset();
        _activity->start();
        std::this_thread::sleep_for(seconds(5));
        _activity->stop();
    }
};

class SlidingVelocityScroller : public ExperimentalVelocityScroller {
private:
    std::unique_ptr<ScrollingActivity> _activity;
    int _velocity;
public:
    SlidingVelocityScroller(ScrollingDevice &dev, int velocity) : ExperimentalVelocityScroller
    (dev), _velocity(velocity) {}
    
    void operator()() {
        _activity = std::unique_ptr<ScrollingActivity>(new SlidingVelocityScrolling(this->device(), _velocity));
        this->reset();
        _activity->start();
        std::this_thread::sleep_for(seconds(5));
        _activity->stop();
    }
};

class ClutchScroller : public ExperimentalVelocityScroller {
private:
    std::unique_ptr<ScrollingActivity> _activity;
public:
    ClutchScroller(ScrollingDevice &dev) 
        : ExperimentalVelocityScroller(dev) {}
    
    void operator()() {
        _activity = std::unique_ptr<ScrollingActivity>(new ClutchingActivity(this->device(), 
                                                                             milliseconds(400), 
                                                                             milliseconds(32), 
                                                                             milliseconds(24), 
                                                                             milliseconds(16)));
        this->reset();
        _activity->start();
        std::this_thread::sleep_for(seconds(5));
        _activity->stop();
        return;
    }
};

class DirectionScroller : public ExperimentalVelocityScroller {
private:
    std::unique_ptr<ScrollingActivity> _activity;
public:
    DirectionScroller(ScrollingDevice &dev) 
        : ExperimentalVelocityScroller(dev) {}

    void operator()() {
        this->reset();
        // Down
        _activity = std::unique_ptr<ScrollingActivity>(new ConstantScrolling(this->device(), -1, milliseconds(8)));
        _activity->start();
        std::this_thread::sleep_for(seconds(2));
        _activity->stop();

        // Up
        _activity = std::unique_ptr<ScrollingActivity>(new ConstantScrolling(this->device(), 1, milliseconds(8)));
        _activity->start();
        std::this_thread::sleep_for(seconds(2));
        _activity->stop();

        // Down
        _activity = std::unique_ptr<ScrollingActivity>(new ConstantScrolling(this->device(), -1, milliseconds(8)));
        _activity->start();
        std::this_thread::sleep_for(seconds(2));
        _activity->stop();
    }
};

#endif
