/*
 *
 * Copyright 2012-2013 Philip Quinn. All Rights Reserved.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */
#include "stdafx.h"
#include <fcntl.h>
#include <io.h>
#include <iostream>
#include "hidapi.h"
#include "devices.hpp"
#include "activities.hpp"
#include "experimental.hpp"

static TCHAR szTitle[] = _T("Scroll Gain Inspector");
static UINT wheelScrollLines = 3; // SPI_GETWHEELSCROLLLINES
static ScrollingDevice *device = nullptr;
static ExperimentalScroller *experiment = nullptr;

void StartScrolling() {
    device = new MicrosoftWheel();
    if (!device->open()) {
        MessageBox(NULL, _T("Error opening EchoMouse."), szTitle, NULL);
        return;
    }

    experiment = new VelocityScroller(*device, 1, 100);
    //experiment = new ClutchScroller(*device);
    //experiment = new MaintainVelocityScroller(*device, 20);
    //experiment = new DirectionScroller(*device);
    experiment->run();
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    short wdelta;

    switch (message) {
    case WM_COMMAND:
        switch (wParam) {
        case BN_CLICKED:
            if (!experiment) {
                StartScrolling();
            } else {
                MessageBox(NULL, _T("Scrolling has already started."), szTitle, NULL);
            }
        }
        break;
    case WM_MOUSEWHEEL:
        wdelta = GET_WHEEL_DELTA_WPARAM(wParam);
        if (experiment) {
            float lines = (wdelta / (float)WHEEL_DELTA) * wheelScrollLines;
            experiment->recordScroll(lines);
        }
        break;
    case WM_SETTINGCHANGE:
        if (wParam == SPI_SETWHEELSCROLLLINES) {
            SystemParametersInfo(SPI_GETWHEELSCROLLLINES, 0, &wheelScrollLines, 0);
        }
        break;
    case WM_DESTROY:
        PostQuitMessage(0);
        break;
    default:
        return DefWindowProc(hWnd, message, wParam, lParam);
        break;
    }

    return 0;
}

void RedirectIOToConsole()
{
    AllocConsole();

    HANDLE hStdOut = GetStdHandle(STD_OUTPUT_HANDLE);
    HANDLE hStdErr = GetStdHandle(STD_ERROR_HANDLE);
    FILE *fp = nullptr;

    CONSOLE_SCREEN_BUFFER_INFO conInfo;
    GetConsoleScreenBufferInfo(hStdOut, &conInfo);
    conInfo.dwSize.Y = 500;
    SetConsoleScreenBufferSize(hStdOut, conInfo.dwSize);

    fp = _fdopen(_open_osfhandle((intptr_t)hStdOut, _O_TEXT), "w");
    *stdout = *fp;
    setvbuf(stdout, NULL, _IONBF, 0);

    fp = _fdopen(_open_osfhandle((intptr_t)hStdErr, _O_TEXT), "w");
    *stderr = *fp;
    setvbuf(stderr, NULL, _IONBF, 0);

    std::ios::sync_with_stdio();
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow) {
    UNREFERENCED_PARAMETER(lpCmdLine);
    UNREFERENCED_PARAMETER(hPrevInstance);

    static TCHAR szWindowClass[] = _T("win32app");

    RedirectIOToConsole();

    if (!SystemParametersInfo(SPI_GETWHEELSCROLLLINES, 0, &wheelScrollLines, 0)) {
        MessageBox(NULL, _T("Error fetching SPI_GETWHEELSCROLLLINES."), szTitle, NULL);
        return 1;
    }

    WNDCLASSEX wcex;
    wcex.cbSize = sizeof(WNDCLASSEX);
    wcex.style          = CS_HREDRAW | CS_VREDRAW;
    wcex.lpfnWndProc    = WndProc;
    wcex.cbClsExtra     = 0;
    wcex.cbWndExtra     = 0;
    wcex.hInstance      = hInstance;
    wcex.hIcon          = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_APPLICATION));
    wcex.hCursor        = LoadCursor(NULL, IDC_ARROW);
    wcex.hbrBackground  = (HBRUSH)(COLOR_WINDOW+1);
    wcex.lpszMenuName   = NULL;
    wcex.lpszClassName  = szWindowClass;
    wcex.hIconSm        = LoadIcon(wcex.hInstance, MAKEINTRESOURCE(IDI_APPLICATION));

    if (!RegisterClassEx(&wcex)) {
        MessageBox(NULL, _T("Call to RegisterClassEx failed!"), szTitle, NULL);
        return 1;
    }

    HWND hWnd = CreateWindow(szWindowClass, szTitle, WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, CW_USEDEFAULT,
        500, 250, NULL, NULL, hInstance, NULL);
    if (!hWnd) {
        MessageBox(NULL, _T("Call to CreateWindow failed!"), szTitle, NULL);
        return 1;
    }

    CreateWindow(TEXT("button"), TEXT("Start Scrolling"), WS_VISIBLE | WS_CHILD | BS_PUSHBUTTON,
        10, 10, 460, 25, hWnd, 0, NULL, NULL);

    ShowWindow(hWnd, nCmdShow);
    UpdateWindow(hWnd);

    // Ensure Sleep() can be called with the greatest possible resolution
    TIMECAPS timerRes;
    if (timeGetDevCaps(&timerRes, sizeof(timerRes)) != MMSYSERR_NOERROR) {
        MessageBox(NULL, _T("Error getting timer resolution information."), szTitle, NULL);
        return -1;
    }
    if (timeBeginPeriod(timerRes.wPeriodMin) != TIMERR_NOERROR) {
        MessageBox(NULL, _T("Error adjusting timer resolution."), szTitle, NULL);
        return -1;
    }

    // Raise process priority
    if (!SetPriorityClass(GetCurrentProcess(), REALTIME_PRIORITY_CLASS)) {
        MessageBox(NULL, _T("Error increasing process priority."), szTitle, NULL);
        return -1;
    }

    // Main message loop:
    MSG msg;
    while (GetMessage(&msg, NULL, 0, 0)) {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }

    if (timeEndPeriod(timerRes.wPeriodMin) != TIMERR_NOERROR) {
        MessageBox(NULL, _T("Error adjusting timer resolution."), szTitle, NULL);
        return -1;
    }

    return (int)msg.wParam;
}
